#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <gmp.h>
#include "librsa.h"

typedef struct {
  mpz_t p, q, e, dP, dQ, qInv;
}
RSA_CRT_private;

typedef struct {
  mpz_t n;
  mpz_t e;
}
RSA_public;

typedef struct {
  uint *ev;
  mpz_t x, s;
}
dpair;

typedef struct {
  mpz_t x, y;
}
rel;

typedef unsigned int uint;
typedef unsigned long ulong;
typedef enum {false, true} bool;

gmp_randstate_t prng;
mpz_t one, two;
ulong max;

void init_prng()
{
  gmp_randinit_default(prng);
  char useed[8193];
  mpz_t seed;

  FILE* fd;
  fd = fopen("/dev/urandom","r");
  if(fd != NULL)
  {
    fread(useed, 8, 1024, fd);
    fclose(fd);
  }

  mpz_init(seed);
  mpz_import(seed, 1024, 1, 8, 0, 0, useed);
  gmp_randseed(prng, seed);
  mpz_clear(seed);

  mpz_init_set_si(one, 1);
  mpz_init_set_si(two, 2);
}

bool is_prime(mpz_t n, bool dp)
{
  int i, j;

  for(i=0; i<NUMPRIMES; i++)
    if(mpz_divisible_ui_p(n, primes[i]))
      return (mpz_cmp_ui(n, primes[i]) ? false : true);

  dp && printf(".") && fflush(stdout);
  mpz_t m, d, a;
  mpz_init_set_si(m, -1); mpz_init(a);
  mpz_add(m,n,m);
  mpz_init_set(d,m);

  ulong s = mpz_remove(d, m, two);
  i=-1;

  loop:
  i++; dp && i>0 && printf("+") && fflush(stdout);
  while(i<miller_rabin_calls)
  {
    mpz_urandomm(a, prng, m);
    mpz_add(a, a, one);

    mpz_powm(a, a, d, n);
    if(mpz_cmp(a, one)==0 || mpz_cmp(a, m)==0) goto loop;

    for(j=1; j<s; j++)
    {
      mpz_powm(a, a, two, n);
      if(mpz_cmp(a, one)==0) goto composite;
      if(mpz_cmp(a, m)==0) goto loop;
    }
    composite:
    mpz_clear(a); mpz_clear(d); mpz_clear(m);
    return false;
  }
  mpz_clear(a); mpz_clear(d); mpz_clear(m);
  return true;
}

void gen_keys(uint size, mpz_t exp, RSA_CRT_private *priv, RSA_public *pub)
{
  mpz_t p,q,n;
  mpz_init(p); mpz_init(q); mpz_init(n);
  mpz_init_set(pub->e, exp);
  mpz_init_set(priv->e, exp);

  gmp_printf("Generating %u bits key pair, using exponent %#Zx (weight=%lu)\n", size, exp, mpz_popcount(exp));

gen:
  mpz_urandomb(p, prng, size/2);
  mpz_setbit(p, 0); mpz_setbit(p, size/2-1); mpz_setbit(p, size/2-2);

  while(!is_prime(p, true)) mpz_add(p, p, two);
  printf(" Found p\n");

  mpz_urandomb(q, prng, size/2);
  mpz_setbit(q, 0); mpz_setbit(q, size/2-1); mpz_setbit(q, size/2-2);

  while(!is_prime(q, true)) mpz_add(q, q, two);
  printf(" Found q\n");

  if(mpz_cmp(p,q)==0){ printf("p=q, retrying ...\n"); goto gen;}
  if(mpz_cmp(p,q)<0) mpz_swap(p,q);

  mpz_init_set(priv->p, p);
  mpz_init_set(priv->q, q);
  mpz_mul(n, p, q);
  mpz_init_set(pub->n, n);

  mpz_invert(n, q, p);
  mpz_init_set(priv->qInv, n);
  mpz_sub(p, p, one);
  mpz_sub(q, q, one);
  mpz_invert(p, pub->e, p);
  mpz_invert(q, pub->e, q);
  mpz_init_set(priv->dP, p);
  mpz_init_set(priv->dQ, q);
  mpz_clear(p); mpz_clear(q); mpz_clear(n);
}

int read_privkey(char *filename, RSA_CRT_private *pk)
{
  FILE* fd;
  if((fd = fopen(filename, "r"))==NULL) return -1;
  mpz_init(pk->p); mpz_init(pk->q); mpz_init(pk->e);
  mpz_init(pk->dP); mpz_init(pk->dQ); mpz_init(pk->qInv);

  if(gmp_fscanf(fd, "P=%Zx\nQ=%Zx\nE=%Zx\ndP=%Zx\ndQ=%Zx\nqInv=%Zx\n", pk->p, pk->q, pk->e, pk->dP, pk->dQ, pk->qInv) < 5)
    return -2;
  printf("Checking primes : ");
  if(!is_prime(pk->p, false) || !is_prime(pk->q, false)) return -3;
  printf(" OK\n");
  mpz_t d, n; mpz_init(d); mpz_init(n);
  mpz_mul(d, pk->dP, pk->e);
  mpz_sub(n, pk->p, one);
  if(mpz_congruent_p(d,one,n) == false) return -4;
  mpz_mul(d, pk->dQ, pk->e);
  mpz_sub(n, pk->q, one);
  if(mpz_congruent_p(d,one,n) == false) return -4;
  mpz_mul(d, pk->q, pk->qInv);
  if(mpz_congruent_p(d,one,pk->p) == false) return -5;
  return 0;
}

int read_pubkey(char *filename, RSA_public *pub)
{
  FILE* fd;
  if((fd = fopen(filename, "r"))==NULL) return -1;
  mpz_init(pub->n); mpz_init(pub->e);
  if(gmp_fscanf(fd, "N=%Zx\nE=%Zx\n", pub->n, pub->e) < 2) return -2;
  return 0;
}

int write_privkey(char *filename, RSA_CRT_private *pk)
{
  FILE* fd;
  if((fd = fopen(filename, "w"))==NULL) return -1;
  gmp_fprintf(fd, "P=%Zx\nQ=%Zx\nE=%Zx\ndP=%Zx\ndQ=%Zx\nqInv=%Zx\n", pk->p, pk->q, pk->e, pk->dP, pk->dQ, pk->qInv);
  fclose(fd);
  return 0;
}

int write_pubkey(char *filename, RSA_public *pub)
{
  FILE* fd;
  if((fd = fopen(filename, "w"))==NULL) return -1;
  gmp_fprintf(fd, "N=%Zx\nE=%Zx\n", pub->n,pub->e);
  fclose(fd);
  return 0;
}

void priv_to_pub(RSA_CRT_private *priv, RSA_public *pub)
{
  mpz_t n;
  mpz_init_set(pub->e, priv->e);
  mpz_init(n);
  mpz_mul(n,priv->p, priv->q);
  mpz_init_set(pub->n, n);
  mpz_clear(n);
}

void rsa_encrypt(RSA_public *pub, char *source, size_t len, char *outbuf)
{
  mpz_t m;
  mpz_init(m);
  mpz_import(m, len, 1, 1, 1, 0, source);
  mpz_powm(m, m, pub->e, pub->n);
  mpz_export(outbuf, NULL, -1, 1, 1, 0, m);
  mpz_clear(m);
}

void rsa_decrypt(RSA_CRT_private *priv, char *source, size_t len, char *outbuf, size_t *oc)
{
  mpz_t m1, m2, c;
  mpz_init(m1); mpz_init(m2); mpz_init(c);
  mpz_import(c, len, -1, 1, 1, 0, source);

  mpz_powm(m1, c, priv->dP, priv->p);
  mpz_powm(m2, c, priv->dQ, priv->q);
  mpz_sub(m1, m1, m2); mpz_mul(c, priv->qInv, m1);
  mpz_mod(c, c, priv->p); mpz_mul(c, c, priv->q);
  mpz_add(c, c, m2);

  if(mpz_tstbit(c, 8*(len-1)+1)==1)
    *oc = len-1-(mpz_scan1(c,0)+1)/8;
  else
    *oc = len-1;

  mpz_export(outbuf, NULL, 1, 1, 1, 0, c);
  mpz_clear(c); mpz_clear(m1); mpz_clear(m2);
}

gboolean hcmp(mpz_t *a, mpz_t *b)
{
  return mpz_cmp(*a, *b)==0;
}

guint hash(mpz_t *x)
{
  return mpz_fdiv_ui(*x, max);
}

void clear_table(mpz_t *x, ulong *y)
{
  mpz_clear(*x); free(x); free(y);
}

int short_attack(RSA_public *pub, char *in, size_t len, char *out, int bits, int mlen)
{
  mpz_t *x, m, c, u; ulong i, *j, k=1;
  GHashTable *ht; void* t; bool found = false;

  max = 1<<bits; mpz_init(m); mpz_init(c); mpz_init(u);
  ht = g_hash_table_new(hash, hcmp);
  printf("Building table "); fflush(stdout);
  mpz_powm_ui(u, two, mpz_sizeinbase(pub->n, 2)-8*mlen-9, pub->n);

  for(i=1; i<max; i++)
  {
    if(i>k*(max/10)){ printf("."); fflush(stdout); k++;}
    mpz_set_ui(m, i); mpz_mul(m, m, u);
    mpz_powm(m, m, pub->e, pub->n);
    x = malloc(sizeof(mpz_t));
    j = malloc(sizeof(ulong));
    mpz_init_set(*x, m); *j = i;
    g_hash_table_insert(ht, x, j);
  }

  printf(" OK\nStarting search "); fflush(stdout);
  mpz_import(c, len, -1, 1, 1, 0, in);

  for(i=1, k=1; i<max; i++)
  {
    if(i>k*(max/10)){ printf("."); fflush(stdout); k++;}
    mpz_set_ui(u, i); mpz_invert(u, u, pub->n);
    mpz_powm(u, u, pub->e, pub->n);
    mpz_mul(u, u, c); mpz_mod(u, u, pub->n);

    j = g_hash_table_lookup(ht, &u);
    if(j != NULL){ found = true; break; }
  }

  printf("\n");
  if(found)
  {
    mpz_set_ui(m, i); mpz_mul_ui(m, m, *j);
    mpz_tdiv_q_2exp(m, m, 1);
    mpz_export(out, NULL, 1, 1, 1, 0, m);
    printf("The attack was successful.\n");
  }
  else printf("The attack failed. Check the length or increate bound.\n");

  g_hash_table_foreach(ht, clear_table, NULL);
  g_hash_table_destroy(ht);
  mpz_clear(m); mpz_clear(c); mpz_clear(u);
  return found;
}

void testprogress(mpz_t n)
{
  ulong i, s; mpz_t u; mpz_init(u); bool done = false;
  for(i=0; i<NUMPRIMES; i++)
  {
    mpz_set_ui(u, primes[i]);
    s = mpz_remove(n, n, u);
    if(s) printf("Found factor %lu^%lu\n", primes[i], s);
  }
  if(mpz_cmp(n,one)==0 || (is_prime(n, false) && gmp_printf("Found factor %Zu\n",n))) done=true;
  mpz_clear(u);
  if(done){mpz_clear(n); exit(0);}
}

bool basef(mpz_t n, ulong *ps, ulong base, dpair *e)
{
  mpz_t v; mpz_init(v);
  ulong i; bool success = false;
  memset(e->ev, 0, sizeof(uint)*base);

  for(i=0; i<base; i++)
  {
    mpz_set_ui(v, ps[i]);
    if(mpz_divisible_p(n,v)) e->ev[i] = mpz_remove(n, n, v);
//    else e->ev[i]=0;
    if(mpz_cmp(n,one)==0) break;
  }
  mpz_clear(v);
  return mpz_cmp_ui(n, 1)==0;
}
/*
int get_pair(mpz_t n, mpz_t sn, ulong *ps, ulong base, dpair *e)
{
  mpz_t u,v; mpz_init(u); mpz_init(v);
  genpair: mpz_urandomm(u, prng, sn); mpz_add(u, u, sn);
  if(mpz_cmp_ui(u,0)==0 || mpz_cmp_ui(u,1)==0) goto genpair;
  mpz_gcd(v,u,n);
  if(mpz_cmp(v, one)>0){
    gmp_printf("Found factor %Zu\n",v); mpz_divexact(n,n,v);
    testprogress(n); return 1;
  }
  mpz_set(e->x, u); mpz_powm(v, u, two, n); mpz_set(e->s, v);
  if(basef(v, ps, base, e)==0) goto genpair;
  mpz_clear(u); mpz_clear(v);
  return 0;
}
*/

void get_pair(mpz_t n, ulong *ps, ulong base, dpair *e, mpz_t sn)
{
  mpz_t w; mpz_init(w);

  do {
    do {
      mpz_set(e->x, sn); mpz_mul(e->s, sn, sn);
      mpz_mod(e->s, e->s, n); mpz_set(w, e->s);
      mpz_add_ui(sn, sn, 1);
    }
    while (mpz_cmp_ui(w, 0) == 0);
  }
  while (basef(w, ps, base, e) == 0);
  mpz_clear(w);
}

int examine_relations(rel *r, mpz_t n, mpz_t factor)
{
  mpz_t sqrt_y;
  mpz_init(sqrt_y);

  if (mpz_perfect_square_p(r->y))
    mpz_sqrt(sqrt_y, r->y);
  else 
    return 0;

  mpz_sub(factor, r->x, sqrt_y);
  mpz_gcd(factor, factor, n);
  mpz_clear(sqrt_y);

  return ((mpz_cmp_ui(factor, 1) != 0) && (mpz_cmp(factor, n) != 0));
}

void form_relations(uint **matrix, uint cols, rel *r, dpair *e, ulong *ps)
{
  uint i, j; mpz_t x;
  mpz_init(x);

  for (i = 0; i < RELATIONS; i++)
  {
    for (j = 0; j < cols+RELATIONS; j++)
    {
      if (matrix[cols+i][j] == 1)
      {
        mpz_mul(r[i].y, r[i].y, e[j].s);
        mpz_mul(r[i].x, r[i].x, e[j].x);
      }
    }
  }
  mpz_clear(x);
}

void gaussian_elim(mpz_t n, dpair *e, rel *r, ulong base, ulong *ps)
{
  uint **matrix, **augment, i, j, k, t;
  int row_elim, col_elim, nonzero_col, nonzero_row;
  int *row1, available_row = 0;

  matrix = malloc(sizeof(uint) * (base + RELATIONS));
  augment = malloc(sizeof(uint) * (base + RELATIONS));

  for (i = 0; i < base + RELATIONS; i ++)
  {
    matrix[i] = malloc(sizeof(uint) * base);
    memset(matrix[i], 0, sizeof(uint) * base);
    augment[i] = malloc(sizeof(uint) * (base + RELATIONS));
    memset(augment[i], 0, sizeof(uint) * (base + RELATIONS));
    augment[i][i] = 1;

    for (j = 0; j < base; j ++)
    {
      matrix[i][j] = e[i].ev[j]%2;
    }
  }

  for (i = 0; i < base; i ++)
  {
    for (j = available_row; j < base + RELATIONS; j++)
    {
      if (matrix[j][i] != 0) 
      {
        nonzero_col = i; nonzero_row = j;
        row1 = matrix[nonzero_row];
        matrix[nonzero_row] = matrix[available_row];
        matrix[available_row] = row1;
        row1 = augment[nonzero_row];
        augment[nonzero_row] = augment[available_row];
        augment[available_row] = row1;

        for (k = nonzero_row + 1; k < base + RELATIONS; k++)
        {
          if (matrix[k][nonzero_col] != 0) 
          {
            for (t = 0; t < base; t ++)
            {
              matrix[k][t] = matrix[k][t] ^ matrix[available_row][t];
            }
            for (t = 0; t < base + RELATIONS; t ++)
            {
              augment[k][t] = augment[k][t] ^ augment[available_row][t];
            }
          }
        }
        available_row ++; j = base + RELATIONS;
      }
    }
  }
  form_relations(augment, base, r, e, ps);
}

void dixion(mpz_t n, ulong base)
{
  uint i, j, k;
  ulong s, *ps; bool prog;
  mpz_t u, sqrtn, factor;
  mpz_init(u); mpz_init(sqrtn);
  dpair texp, *tv; rel *r;
  testprogress(n);

  ps = malloc(base*sizeof(ulong));
  for(i=2,s=0,mpz_set(u,two); s<base; i++)
  {
    if(is_prime(u,false)) ps[s++] = i;
    mpz_add(u,u,one);
  }
  gmp_printf("Loaded a table of %lu primes for N=%Zu\n", base,n);
  mpz_sqrt(sqrtn, n);

  texp.ev = malloc(sizeof(uint) * base);
  memset(texp.ev, 0, sizeof(uint) * base);
  mpz_init(texp.x); mpz_init(texp.s);

  tv = malloc(sizeof(dpair) * (base + RELATIONS));
  for (i = 0; i < base + RELATIONS; i++)
  {
    tv[i].ev = malloc(sizeof(uint) * base);
//    memset(tv[i].ev, 0, sizeof(uint) * base);
    mpz_init(tv[i].x); mpz_init(tv[i].s);
  }

  printf("Looking for pairs x = y^2 mod n (need to find %lu) :\n", base+RELATIONS);
start: prog = false;

  for (i = 0; i < base + RELATIONS; i++)
  {
    printf("."); fflush(stdout);
    get_pair(n, ps, base, &tv[i], sqrtn);
/*
  gmp_printf("\nFound pair %Zu, %Zu\n", tv[i].x, tv[i].s);
  for (j = 0; j < base; j++) printf("%d ", tv[i].ev[j]);
  printf("\n");
*/
  } printf(" OK\n");

  r = malloc(sizeof(rel) * RELATIONS);
  for (i = 0; i < RELATIONS; i++)
  {
    mpz_init_set_ui(r[i].x, 1);
    mpz_init_set_ui(r[i].y, 1);
  }
  gaussian_elim(n, tv, r, base, ps);

  mpz_init(factor);
  for (i = 0; i < RELATIONS; i++)
  {
//    gmp_printf("Testing relation %Zu -> %Zu\n", r[i].x, r[i].y);
    if (examine_relations(&r[i], n, factor))
    {
      mpz_mod(u, n, factor);
      if (mpz_cmp_ui(u, 0) == 0)
      {
        gmp_printf("Found factor %Zu\n", factor);
        mpz_div(n, n, factor);
        testprogress(n); prog = true;
      }
    }
  }
  if(prog) goto start;
  gmp_printf("Fail : did not find any factor of %Zu\n", n);

  end:
  mpz_clear(u); mpz_clear(factor); free(ps);
}
