#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <string.h>
#include <math.h>
#include "librsa.c"

char *pname;

void usage()
{
  printf("Usage : %s <command> [options]\n\n", pname);
  printf("This is a basic implementation of the RSA Cryptosystem.\n");
  printf("You can generate a new pair of keys using the 'gen' command\n");
  printf("  Possible options for this command are :\n");
  printf("    -n size : set the modulus size in bits\n");
  printf("    -k tries : set number of bases to test in the Miller-Rabin test\n");
  printf("    -e exponent : set the public exponent (default is 0x10001)\n");
  printf("    -o file : file in which to store the private key\n");
  printf("    -p file : file in which to store the public key\n");
  printf("\nYou can test your private key with the 'key' command\n");
  printf("  Possible options for this command are :\n");
  printf("    -i file : private key file (required)\n");
  printf("    -k tries : set number of bases to test in the Miller-Rabin test\n");
  printf("    -p file : write the associated public key to file\n");
  printf("    -v : print details about the key\n");
  printf("\nYou can encrypt a file with the 'crypt' command\n");
  printf("  Required options for this command are :\n");
  printf("    -i file : file to encrypt\n");
  printf("    -o file : encrypted file name\n");
  printf("    -p file : public key file\n");
  printf("\nYou can decrypt a file with the 'decrypt' command\n");
  printf("  Required options for this command are :\n");
  printf("    -i file : crypted file\n");
  printf("    -o file : decrypted file name\n");
  printf("    -p file : private key file\n");
  printf("\nYou can run a short message attack with the 'short' command\n");
  printf("  Required options for this command are :\n");
  printf("    -i file : crypted file\n");
  printf("    -o file : output file\n");
  printf("    -p file : public key file\n");
  printf("    -l length : length in bytes of the unpadded clear text\n");
  printf("    -b bound : (optional) boud in bits for cypher factors\n");
  printf("\nA terrible implementation of Dexion's algorithm is available using the 'fact' command\n");
  printf(" Do not attempt to factorize any RSA modulus larger than 100 bits\n");
  printf("  Possible options for this command are :\n");
  printf("    -n number (required). Use 0x to input in hexadecimal\n");
  printf("    -b base : number of primes in the base\n");
  printf("    -r relations : number of x=y^2 relations\n");
  printf("\nWritten by Antoine Delignat-Lavaud <antoine.delignat-lavaud@ens-cachan.fr>\n");
  exit(0);
}

void clear()
{
  mpz_clear(one); mpz_clear(two); gmp_randclear(prng);
}

int main(int argc, char **argv)
{
  int i, j;
  RSA_public pub;
  RSA_CRT_private priv;
  char* cmd;
  char* privkfile;
  char* pubkfile;
  char* outfile;

  pname = argv[0];
  if(argc < 2) usage();
  init_prng(); atexit(clear);

  cmd = argv[1];
  switch(cmd[0])
  {
    // Generate keys
    case 'g':
    {
      int keylen = 1024;
      bool outfile = false;
      bool outpub = false;
      mpz_t e; mpz_init_set_ui(e, 0x10001L);

      for(i=2; i<argc; i++)
      {
        if(argv[i][0]=='-')
        {
          switch(argv[i][1])
          {
            case 'n': (++i < argc) && (keylen = atoi(argv[i]));
              if(keylen <= 8){printf("Error : invalid key size %d !\n", keylen); usage();}
              break;
            case 'k': ++i < argc && (miller_rabin_calls = atoi(argv[i]));
              if(miller_rabin_calls < 3){
                printf("Error : you need at least 3 Miller-Rabin calls, %d given\n", miller_rabin_calls);
                usage();
              } break;
             case 'e': if(mpz_set_str(e, argv[++i],0))
                {printf("Error : %s is not a valid exponent\n", argv[i]); usage();}
                if(mpz_cmp_si(e,2)<=0)
                {printf("Error : exponent should be at least 3\n"); usage();}
                break;
             case 'o': ++i < argc && (privkfile = argv[i]) && (outfile = true);
                break;
             case 'p': ++i < argc && (pubkfile = argv[i]) && (outpub = true);
                break;
             default: printf("Unknown option %s\n", argv[i]); usage();
          }
        }
        else
        {
          printf("Error : unknown argument %s\n", argv[i]);
          usage();
        }
      }
      gen_keys(keylen, e, &priv, &pub);

      if(outfile)
      {
        i = write_privkey(privkfile, &priv);
        switch(i){ case 0: printf("Private key written to %s.\n", privkfile); break;
          case -1: printf("Error : can't open %s for writing !\n", privkfile); break;
        }
      }
      else
      {
        gmp_printf("P = %#Zx\nQ = %#Zx\nE = %#Zx\nN = %#Zx\ndP = %#Zx\ndQ = %#Zx\nqInv = %#Zx\n",
                    priv.p, priv.q, priv.e, pub.n, priv.dP, priv.dQ, priv.qInv);
      }

      if(outpub)
      {
        i = write_pubkey(pubkfile, &pub);
        switch(i){ case 0: printf("Public key written to %s.\n", pubkfile); break;
          case -1: printf("Error : can't open %s for writing !\n", pubkfile); break;
        }
      }
      }
      break;

    // Test key
    case 'k':
    {
      bool verbose = false;
      bool privk = false;
      bool pubk = false;

      for(i=2; i<argc; i++)
      {
        if(argv[i][0]=='-')
        {
          switch(argv[i][1])
          {
            case 'v': verbose = true; break;
            case 'k': ++i < argc && (miller_rabin_calls = atoi(argv[i]));
              if(miller_rabin_calls < 4){
                printf("Error : you need at least 4 Miller-Rabin calls, %d given\n", miller_rabin_calls);
                usage();
              } break;
             case 'i': ++i < argc && (privkfile = argv[i]) && (privk = true);
                break;
             case 'p': ++i < argc && (pubkfile = argv[i]) && (pubk = true);
                break;
             default: printf("Unknown option %s\n", argv[i]); usage();
          }
        }
        else
        {
          printf("Error : unknown argument %s\n", argv[i]);
          usage();
        }
      }
      if(!privk){printf("Error : no private key file provided.\n"); usage();}

      i = read_privkey(privkfile, &priv);
      switch(i)
      { case 0: printf("Succesfully loaded private key %s\n", privkfile); break;
        case -1: printf("Can't open file %s\n", privkfile); break;
        case -2: printf("%s is not a valid private key file\n", privkfile); break;
        case -3: printf("Error : private key doesn't have primes\n"); break;
        case -4: printf("Invalid private exponents in private key !\n"); break;
        case -5: printf("Invalid CRT coefficient in private key !\n"); break;
      } if(i<0) exit(1);

      if(verbose)
      {
        gmp_printf("P = %#Zx\nQ = %#Zx\nE = %#Zx\ndP = %#Zx\ndQ = %#Zx\nqInv = %#Zx\n", priv.p, priv.q, priv.e, priv.dP, priv.dQ, priv.qInv); 
      }

      if(pubk)
      {
        priv_to_pub(&priv, &pub);
        i = write_pubkey(pubkfile, &pub);
        switch(i){ case 0: printf("Public key written to %s.\n", pubkfile); break;
          case -1: printf("Error : can't open %s for writing !\n", pubkfile); break;
        }
      }
      }
      break;

    // Encryption
    case 'c':
    {
      bool pubf = false;
      bool in = false;
      bool out = false;

      for(i=2; i<argc; i++)
      {
        if(argv[i][0]=='-')
        {
          switch(argv[i][1])
          {
             case 'o': ++i < argc && (outfile = argv[i]) && (out = true);
                break;
             case 'i': ++i < argc && (privkfile = argv[i]) && (in = true);
                break;
             case 'p': ++i < argc && (pubkfile = argv[i]) && (pubf = true);
                break;
             default: printf("Unknown option %s\n", argv[i]); usage();
          }
        }
        else
        {
          printf("Error : unknown argument %s\n", argv[i]);
          usage();
        }
      }
      if(!pubf){printf("Error : public key missing.\n"); usage();}
      if(!in){printf("Error : input file missing.\n"); usage();}
      if(!out){printf("Error : output file missing.\n"); usage();}

      switch(read_pubkey(pubkfile, &pub)){
        case -1: printf("Can't open public key file %s\n",pubkfile); usage();
        case -2: printf("%s is not a valid public key file !\n", pubkfile); usage();
      }

      size_t mod_len = mpz_sizeinbase(pub.n, 2);
      if(mod_len <= 8) return printf("Error : the modulus is way too short !\n");

      FILE* fin;
      FILE* fout;

      if((fin=fopen(privkfile,"r"))==NULL) return printf("Error : can't open %s for reading !\n",privkfile);
      if((fout=fopen(outfile,"w"))==NULL) return printf("Error : can't open %s for writing !\n",outfile);
      printf("Warning: using a weak static padding to allow cryptanalysis\nBlocks : ");

      size_t len, block_len = (mod_len-1)/8, c_len;
      c_len = mod_len/8 + (mod_len%8 != 0);

      char *inbuf = malloc(block_len+1); inbuf[0]=1;
      char *outbuf = malloc(c_len);
      while(len = fread(inbuf+1, 1, block_len, fin))
      {
        printf("#"); fflush(stdout);
        if(len < block_len){ inbuf[0]=2; inbuf[++len]=128; memset(inbuf+len+1,0,block_len-len);}
        memset(outbuf, 0, c_len);
        rsa_encrypt(&pub, inbuf, block_len+1, outbuf);
        fwrite(outbuf, 1, c_len, fout);
      }
      printf("\n");
      fclose(fin); fclose(fout);
      free(inbuf); free(outbuf);
      }
      break;

    // Decryption
    case 'd':
    {
      bool privf = false;
      bool in = false;
      bool out = false;

      for(i=2; i<argc; i++)
      {
        if(argv[i][0]=='-')
        {
          switch(argv[i][1])
          {
             case 'o': ++i < argc && (outfile = argv[i]) && (out = true);
                break;
             case 'i': ++i < argc && (pubkfile = argv[i]) && (in = true);
                break;
             case 'p': ++i < argc && (privkfile = argv[i]) && (privf = true);
                break;
             default: printf("Unknown option %s\n", argv[i]); usage();
          }
        }
        else
        {
          printf("Error : unknown argument %s\n", argv[i]);
          usage();
        }
      }
      if(!privf){printf("Error : private key missing.\n"); usage();}
      if(!in){printf("Error : input file missing.\n"); usage();}
      if(!out){printf("Error : output file missing.\n"); usage();}

      i = read_privkey(privkfile, &priv);
      switch(i)
      { case 0: printf("Succesfully loaded private key %s\n", privkfile); break;
        case -1: printf("Can't open file %s\n", privkfile); break;
        case -2: printf("%s is not a valid private key file\n", privkfile); break;
        case -3: printf("Error : private key doesn't have primes\n"); break;
        case -4: printf("Invalid private exponents in private key !\n"); break;
        case -5: printf("Invalid CRT coefficient in private key !\n"); break;
      } if(i<0) exit(1);

      priv_to_pub(&priv, &pub);
      size_t mod_len = mpz_sizeinbase(pub.n, 2);

      FILE* fin;
      FILE* fout;

      if((fin=fopen(pubkfile,"r"))==NULL) return printf("Error : can't open %s for reading !\n",privkfile);
      if((fout=fopen(outfile,"w"))==NULL) return printf("Error : can't open %s for writing !\n",outfile);

      size_t len, oc, c_len = mod_len/8 + (mod_len%8==0? 0 : 1);
      char *inbuf = malloc(c_len);
      char *outbuf = malloc(c_len);
      while(len = fread(inbuf, 1, c_len, fin))
      {
        memset(outbuf, 0, c_len); printf("#"); fflush(stdout);
        rsa_decrypt(&priv, inbuf, len, outbuf, &oc);
        fwrite(outbuf+1, 1, oc, fout);
      }
      printf("\n");
      fclose(fin); fclose(fout);
      free(inbuf); free(outbuf);
      }
      break;


    // Attack on short messages
    case 's':
    {
      bool pubf = false;
      bool in = false;
      bool out = false;
      bool cl = false;
      int bits = 0, mlen;

      for(i=2; i<argc; i++)
      {
        if(argv[i][0]=='-')
        {
          switch(argv[i][1])
          {
             case 'o': ++i < argc && (outfile = argv[i]) && (out = true);
                break;
             case 'i': ++i < argc && (privkfile = argv[i]) && (in = true);
                break;
             case 'p': ++i < argc && (pubkfile = argv[i]) && (pubf = true);
                break;
            case 'l': (++i < argc) && (mlen = atoi(argv[i])) && (cl = true);
              if(mlen < 1){printf("Invalid cleartext length : %d\n", mlen); usage();}
              break;
            case 'b': (++i < argc) && (bits = atoi(argv[i]));
              if(bits < 8){printf("Error : not enough bits to factorize an ASCII string\n", bits); usage();}
              break;
             default: printf("Unknown option %s\n", argv[i]); usage();
          }
        }
        else
        {
          printf("Error : unknown argument %s\n", argv[i]);
          usage();
        }
      }
      if(!pubf){printf("Error : public key missing.\n"); usage();}
      if(!in){printf("Error : input file missing.\n"); usage();}
      if(!out){printf("Error : output file missing.\n"); usage();}
      if(!cl){printf("Error : cleartext length missing (required because of 0-padding)\n"); usage();}
      if(bits && bits < 4*mlen) return printf("Error : you will need at least a %d bits boundary !", 4*mlen);
      else bits = 4*mlen+2;

      switch(read_pubkey(pubkfile, &pub)){
        case -1: printf("Can't open public key file %s\n",pubkfile); usage();
        case -2: printf("%s is not a valid public key file !\n", pubkfile); usage();
      }

      size_t mod_len = mpz_sizeinbase(pub.n, 2);
      if(mod_len < 64) return printf("It will be faster to factor the %d bits modulus.\n", mod_len);
      size_t c_len = mod_len/8 + (mod_len%8==0? 0 : 1);

      FILE *fin, *fout;
      if((fin=fopen(privkfile,"r"))==NULL) return printf("Error : can't open %s for reading !\n",privkfile);

      char *inbuf = malloc(c_len);
      char *outbuf = malloc(c_len);
      size_t len = fread(inbuf, 1, c_len+2, fin);
      if(len > c_len) return printf("Error : message has several blocks. You may try the attack the last block.\n");

      printf("Starting the attack, l=%d, b=%d. %d bits must be discovered\n", mlen, bits, 8*mlen+2);
      if(short_attack(&pub, inbuf, len, outbuf, bits, mlen))
      {
        if((fout=fopen(outfile,"w"))==NULL) return printf("Error : can't open %s for writing !\n",outfile);
        fwrite(outbuf+1, 1, mlen, fout);
        fclose(fout);
      }

      fclose(fin);
      free(inbuf); free(outbuf);
      }
      break;

    // Factorization
    case 'f':
    {
      mpz_t n; mpz_init_set_ui(n, 0);
      bool fact = false;
      ulong bound = 200;

      for(i=2; i<argc; i++)
      {
        if(argv[i][0]=='-')
        {
          switch(argv[i][1])
          {
             case 'n': (++i < argc) && (fact = true) && mpz_set_str(n, argv[i], 0);
                if(mpz_cmp_ui(n, 1)<=0)
                {printf("Expecting a positive integer greater than 2\n"); usage();}
                break;
             case 'b': (++i < argc) && (bound = atoi(argv[i]));
               if(bound < 3){printf("Invalid base : %lu\n", bound); usage();}
               break;
             case 'r': (++i < argc) && (RELATIONS = atoi(argv[i]));
               if(RELATIONS < 2){printf("Invalid number of relations : %lu\n", RELATIONS); usage();}
               break;
             default: printf("Unknown option %s\n", argv[i]); usage();
          }
        }
        else
        {
          printf("Error : unknown argument %s\n", argv[i]);
          usage();
        }
      }
      if(!fact){ printf("Error : no number to factorize.\n"); usage();}

      printf("Warning: due to a poor implementation this will likely eat up all your RAM.\n");
      dixion(n, bound);
      mpz_clear(n);
      }
      break;

      default: printf("Unknown command %s\n",argv[1]);
  }

  return 0;
}
